/************************************************************************************


Peer-to-peer wireless communication for GRBL using ESP_NOW.

This application implements a peer-to-peer bidirectional communication stream for use with GRBL.
It is used to connect a GRBL machine such as a CNC mill or 3D printer with a PC host using
a wireless link that does not require support from a server or WiFi. The link will auto-connect.

It runs on any MCU in the ESP line that supports ESP-NOW. As of 2024 this includes 
ESP8266, ESP32, ESP32-S and ESP32-C.

This application is configured for the ESP8266 and uses only a small subset of the ESP-NOW 
features, so it should run on any of the ESP modules.

*//////////////////////////////////////////////////////////////////////////////////////

#include <ESP8266WiFi.h>
#include <espnow.h>

// Insert the MAC addresses of the two communication devices here.
// This code can be used for either end of the link by commenting/uncommenting these lines.
// Use the MAC Address of the OTHER (receiving) unit.
uint8_t broadcastAddress[] = { 0x2c, 0xf4, 0x32, 0x89, 0xab, 0xdf };
// uint8_t broadcastAddress[] = {0x48, 0x3f, 0xda, 0x03, 0x15, 0xdc};

// ESP-NOW supports a message length max of 250.  GRBL max message length is 80.
#define MAX_MSG_LEN 200

// Define variables to store messages
char inMsg[MAX_MSG_LEN];   // Buffer for receive message
char outMsg[MAX_MSG_LEN];  // Buffer for send message
int pos;                   // Buffer pointer for send message
bool termCR0 = false;      // Flag for Cr or CrLf terminator
bool termLF0 = false;      // Flag for Lf or LfCr terminator
bool termLF1 = false;      // Flag for CrLf terminator
bool termCR1 = false;      // Flag for LfCr terminator

// Variable to store if sending data was successful
String success;

// Callback when data is sent
void OnDataSent(uint8_t *mac_addr, uint8_t sendStatus) {
  // Suppress send log messages in console passthru mode
  // Serial.print("Last Packet Send Status: ");
  if (sendStatus != 0) {
    Serial.print("Delivery Fail ");
  }
}

// Callback when data is received
// Data rceived from the other unit is passed exactly as received.
void OnDataRecv(uint8_t *mac, uint8_t *incomingData, uint8_t len) {
  // Serial.println(len);
  memcpy(&inMsg, incomingData, len);

  // Debug display length and content
  //inMsg[len] = '\0';
  //Serial.print("Bytes received: ");
  //Serial.println(len);
  //Serial.print(" Text: ");
  //printIncomingMessage();

  Serial.write(inMsg);     // Console passthrough

}

void printIncomingMessage() {
  // Display Readings in Serial Monitor
  // Serial.println(inMsg);
  for (uint i = 0; i < sizeof(inMsg); i++) {
    if (inMsg[i] == '\0') break;
    Serial.print(inMsg[i], HEX);
    Serial.print(" ");
  }
  Serial.println();
}

void setup() {
  // Init Serial Monitor
  Serial.begin(115200);
  Serial.println();
  Serial.println("GRBL TRANSFER");

  // Set device as a Wi-Fi Station
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  Serial.println(WiFi.macAddress());

  // Init ESP-NOW
  if (esp_now_init() != 0) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }

  // Set ESP-NOW Role
  esp_now_set_self_role(ESP_NOW_ROLE_COMBO);

  // Register the callback to get the status of transmitted packet
  esp_now_register_send_cb(OnDataSent);

  // Register peer
  esp_now_add_peer(broadcastAddress, ESP_NOW_ROLE_COMBO, 1, NULL, 0);

  // Register for a callback function for when data is received
  esp_now_register_recv_cb(OnDataRecv);
}

void loop() {
  static char c;
  while (Serial.available() > 0) {
    c = Serial.read();
    // Check for immediate (single-character) commands
    if ((c == '\x18') || (c == '?') || (c == '~') || (c == '!') || (c > 127)) {
      doImmediate(c);
      break;  // Get next character
    }

    // Check for message terminator
    if (c == '\r') {   // Message ends in Cr or CrLf
      termCR0 = true;  // Message ends in Cr
      int i, j = 0;
      do {
        i = Serial.peek();  // Loop until next char available
        j++;
      } while ((i == -1) && (j < 55));
      if ((char)i == '\n') {  // If it's a LF
        c = Serial.read();    //  then get it and
        termLF1 = true;       //  Message ends in CrLf
      }
    } else if (c == '\n') {   // Message ends in Lf or LfCr
      termLF0 = true;         // Message ends in Lf
      int i, j = 0;
      do {
        i = Serial.peek();  // Loop until next char available
        j++;
      } while ((i == -1) && (j < 55));
      if ((char)i == '\r') {  // If it's a Cr
        c = Serial.read();    //  then get it and
        termCR1 = true;       //  Message ends in LfCr
      }
    }

    if ((!termCR0) && (!termLF0)) {  //  No endline detected yet
      outMsg[pos] = c;               //   add input to the bufffer and
      pos++;                         //   bump the buffer pointer.
      break;
    }

    // One or both terminators are set.
    // Process buffer to NOW output.

    // Test for $X (keyboard entry) for soft reset
    if ((outMsg[0] == '$') && (outMsg[1] == 'X')) {
      doImmediate('\x18');
      pos = 0;
    }

    // Add terminators as received (Cr, Lf or CrLf, LfCr).
    if (termCR0 == true) outMsg[pos++] = '\r';
    if (termLF0 == true) outMsg[pos++] = '\n';
    if (termCR1 == true) outMsg[pos++] = '\r';
    if (termLF1 == true) outMsg[pos++] = '\n';
    outMsg[pos++] = '\0';  // Insert string terminator
    termCR0 = false;       // Clear terminator flags
    termLF0 = false;
    termCR1 = false;
    termLF1 = false;

    // Debug Show values to send
    // Serial.print(pos);
    // Serial.print(" ");
    // Serial.print(outMsg);

    // Send message via ESP-NOW
    esp_now_send(broadcastAddress, (uint8_t *)&outMsg, pos);

    pos = 0;  // Reset message bufffer pointer
  }           // is.available
}  // Loop

void doImmediate(char c) {
  // Single-character command. No Cr/Lf
  static char immMsg[2];
  immMsg[0] = c;
  immMsg[1] = '\0';
  esp_now_send(broadcastAddress, (uint8_t *)&immMsg, 2);
}
